@@ -1,5 +1,6 @@
 import asyncio
 import base64
+import datetime
 import hashlib
 import hmac
 import logging
@@ -12,6 +13,10 @@ import aiohttp
 from aiohttp.http_websocket import json
 from aiohttp.typedefs import StrOrURL
 
+import pybotters
+
+from .auth import Auth as _Auth
+
 logger = logging.getLogger(__name__)
 
 
@@ -20,18 +25,21 @@ async def ws_run_forever(
     session: aiohttp.ClientSession,
     event: asyncio.Event,
     *,
-    send_str: Optional[Union[str, List[str]]]=None,
-    send_json: Any=None,
+    send_str: Optional[Union[str, List[str]]] = None,
+    send_json: Any = None,
     hdlr_str=None,
     hdlr_json=None,
+    auth=_Auth,
     **kwargs: Any,
 ) -> None:
+    if all([hdlr_str is None, hdlr_json is None]):
+        hdlr_json = pybotters.print_handler
     iscorofunc_str = asyncio.iscoroutinefunction(hdlr_str)
     iscorofunc_json = asyncio.iscoroutinefunction(hdlr_json)
     while not session.closed:
-        separator = asyncio.create_task(asyncio.sleep(60.0))
+        cooldown = asyncio.create_task(asyncio.sleep(60.0))
         try:
-            async with session.ws_connect(url, **kwargs) as ws:
+            async with session.ws_connect(url, auth=auth, **kwargs) as ws:
                 event.set()
                 if '_authtask' in ws.__dict__:
                     await ws.__dict__['_authtask']
@@ -42,7 +50,9 @@ async def ws_run_forever(
                         await ws.send_str(send_str)
                 if send_json is not None:
                     if isinstance(send_json, list):
-                        await asyncio.gather(*[ws.send_json(item) for item in send_json])
+                        await asyncio.gather(
+                            *[ws.send_json(item) for item in send_json]
+                        )
                     else:
                         await ws.send_json(send_json)
                 async for msg in ws:
@@ -70,9 +80,9 @@ async def ws_run_forever(
                                     logger.error(repr(e))
                     elif msg.type == aiohttp.WSMsgType.ERROR:
                         break
-        except aiohttp.WSServerHandshakeError as e:
+        except (aiohttp.WSServerHandshakeError, aiohttp.ClientOSError) as e:
             logger.warning(repr(e))
-        await separator
+        await cooldown
 
 
 class Heartbeat:
@@ -83,10 +93,10 @@ class Heartbeat:
             await asyncio.sleep(30.0)
 
     @staticmethod
-    async def btcmex(ws: aiohttp.ClientWebSocketResponse):
+    async def bitbank(ws: aiohttp.ClientWebSocketResponse):
         while not ws.closed:
-            await ws.send_str('ping')
-            await asyncio.sleep(30.0)
+            await ws.send_str('2')
+            await asyncio.sleep(15.0)
 
     @staticmethod
     async def liquid(ws: aiohttp.ClientWebSocketResponse):
@@ -106,23 +116,40 @@ class Heartbeat:
             await ws.pong()
             await asyncio.sleep(60.0)
 
+    @staticmethod
+    async def phemex(ws: aiohttp.ClientWebSocketResponse):
+        while not ws.closed:
+            await ws.send_str('{"method":"server.ping","params":[],"id":123}')
+            await asyncio.sleep(10.0)
+
 
 class Auth:
     @staticmethod
     async def bitflyer(ws: aiohttp.ClientWebSocketResponse):
-        key: str = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][0]
-        secret: bytes = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][1]
+        key: str = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][0]
+        secret: bytes = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][1]
 
         timestamp = int(time.time())
         nonce = token_hex(16)
-        sign = hmac.new(secret, f'{timestamp}{nonce}'.encode(), digestmod=hashlib.sha256).hexdigest()
-        await ws.send_json({
-            'method': 'auth',
-            'params': {
-                'api_key': key, 'timestamp': timestamp, 'nonce': nonce, 'signature': sign
-            },
-            'id': 'auth',
-        })
+        sign = hmac.new(
+            secret, f'{timestamp}{nonce}'.encode(), digestmod=hashlib.sha256
+        ).hexdigest()
+        await ws.send_json(
+            {
+                'method': 'auth',
+                'params': {
+                    'api_key': key,
+                    'timestamp': timestamp,
+                    'nonce': nonce,
+                    'signature': sign,
+                },
+                'id': 'auth',
+            }
+        )
         async for msg in ws:
             if msg.type == aiohttp.WSMsgType.TEXT:
                 data = msg.json()
@@ -134,11 +161,19 @@ class Auth:
 
     @staticmethod
     async def liquid(ws: aiohttp.ClientWebSocketResponse):
-        key: str = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][0]
-        secret: bytes = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][1]
+        key: str = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][0]
+        secret: bytes = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][1]
 
         json_payload = json.dumps(
-            {'path': '/realtime', 'nonce': str(int(time.time() * 1000)), 'token_id': key},
+            {
+                'path': '/realtime',
+                'nonce': str(int(time.time() * 1000)),
+                'token_id': key,
+            },
             separators=(',', ':'),
         ).encode()
         json_header = json.dumps(
@@ -151,37 +186,70 @@ class Auth:
         ]
         signing_input = b'.'.join(segments)
         signature = hmac.new(secret, signing_input, hashlib.sha256).digest()
-        segments.append(
-            base64.urlsafe_b64encode(signature).replace(b'=', b'')
-        )
+        segments.append(base64.urlsafe_b64encode(signature).replace(b'=', b''))
         encoded_string = b'.'.join(segments).decode()
 
-        await ws.send_json({
-            'event': 'quoine:auth_request',
-            'data': {
-                'path': '/realtime',
-                'headers': {'X-Quoine-Auth': encoded_string},
-            },
-        })
+        await ws.send_json(
+            {
+                'event': 'quoine:auth_request',
+                'data': {
+                    'path': '/realtime',
+                    'headers': {'X-Quoine-Auth': encoded_string},
+                },
+            }
+        )
 
     @staticmethod
     async def ftx(ws: aiohttp.ClientWebSocketResponse):
-        key: str = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][0]
-        secret: bytes = ws._response._session.__dict__['_apis'][AuthHosts.items[ws._response.url.host].name][1]
+        key: str = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][0]
+        secret: bytes = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][1]
 
         ts = int(time.time() * 1000)
-        sign = hmac.new(secret, f'{ts}websocket_login'.encode(), digestmod=hashlib.sha256).hexdigest()
+        sign = hmac.new(
+            secret, f'{ts}websocket_login'.encode(), digestmod=hashlib.sha256
+        ).hexdigest()
 
         msg = {
             'op': 'login',
-            'args': {
-                'key': key, 'sign': sign, 'time': ts
-            },
+            'args': {'key': key, 'sign': sign, 'time': ts},
         }
         if 'FTX-SUBACCOUNT' in ws._response.request_info.headers:
-            msg['args']['subaccount'] = ws._response.request_info.headers['FTX-SUBACCOUNT']
+            msg['args']['subaccount'] = ws._response.request_info.headers[
+                'FTX-SUBACCOUNT'
+            ]
         await ws.send_json(msg)
 
+    @staticmethod
+    async def phemex(ws: aiohttp.ClientWebSocketResponse):
+        key: str = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][0]
+        secret: bytes = ws._response._session.__dict__['_apis'][
+            AuthHosts.items[ws._response.url.host].name
+        ][1]
+
+        expiry = int(time.time() + 60.0)
+        signature = hmac.new(
+            secret, f'{key}{expiry}'.encode(), digestmod=hashlib.sha256
+        ).hexdigest()
+        msg = {
+            'method': 'user.auth',
+            'params': ['API', key, signature, expiry],
+            'id': 123,
+        }
+        await ws.send_json(msg)
+        async for msg in ws:
+            if msg.type == aiohttp.WSMsgType.TEXT:
+                data = msg.json()
+                if data['result'] == {'status': 'success'}:
+                    break
+            elif msg.type == aiohttp.WSMsgType.ERROR:
+                break
+
 
 @dataclass
 class Item:
@@ -191,11 +259,10 @@ class Item:
 
 class HeartbeatHosts:
     items = {
-        'www.btcmex.com': Heartbeat.btcmex,
+        'stream.bitbank.cc': Heartbeat.bitbank,
         'stream.bybit.com': Heartbeat.bybit,
         'stream.bytick.com': Heartbeat.bybit,
         'stream-testnet.bybit.com': Heartbeat.bybit,
-        'stream-testnet.bybit.com': Heartbeat.bybit,
         'tap.liquid.com': Heartbeat.liquid,
         'ftx.com': Heartbeat.ftx,
         'stream.binance.com': Heartbeat.binance,
@@ -206,6 +273,8 @@ class HeartbeatHosts:
         'dstream.binancefuture.com': Heartbeat.binance,
         'testnet.binanceops.com': Heartbeat.binance,
         'testnetws.binanceops.com': Heartbeat.binance,
+        'phemex.com': Heartbeat.phemex,
+        'testnet.phemex.com': Heartbeat.phemex,
     }
 
 
@@ -214,6 +283,8 @@ class AuthHosts:
         'ws.lightstream.bitflyer.com': Item('bitflyer', Auth.bitflyer),
         'tap.liquid.com': Item('liquid', Auth.liquid),
         'ftx.com': Item('ftx', Auth.ftx),
+        'phemex.com': Item('phemex', Auth.phemex),
+        'testnet.phemex.com': Item('phemex_testnet', Auth.phemex),
     }
 
 
@@ -221,7 +292,51 @@ class ClientWebSocketResponse(aiohttp.ClientWebSocketResponse):
     def __init__(self, *args, **kwargs) -> None:
         super().__init__(*args, **kwargs)
         if self._response.url.host in HeartbeatHosts.items:
-            self.__dict__['_pingtask'] = asyncio.create_task(HeartbeatHosts.items[self._response.url.host](self))
-        if self._response.url.host in AuthHosts.items:
-            if AuthHosts.items[self._response.url.host].name in self._response._session.__dict__['_apis']:
-                self.__dict__['_authtask'] = asyncio.create_task(AuthHosts.items[self._response.url.host].func(self))
+            self.__dict__['_pingtask'] = asyncio.create_task(
+                HeartbeatHosts.items[self._response.url.host](self)
+            )
+        if self._response.__dict__['_auth'] is _Auth:
+            if self._response.url.host in AuthHosts.items:
+                if (
+                    AuthHosts.items[self._response.url.host].name
+                    in self._response._session.__dict__['_apis']
+                ):
+                    self.__dict__['_authtask'] = asyncio.create_task(
+                        AuthHosts.items[self._response.url.host].func(self)
+                    )
+        self._lock = asyncio.Lock()
+
+    async def send_str(self, *args, **kwargs) -> None:
+        if self._response.url.host not in RequestLimitHosts.items:
+            await super().send_str(*args, **kwargs)
+        else:
+            super_send_str = super().send_str(*args, **kwargs)
+            await RequestLimitHosts.items[self._response.url.host](self, super_send_str)
+
+
+class RequestLimit:
+    @staticmethod
+    async def gmocoin(ws: ClientWebSocketResponse, send_str):
+        async with ws._lock:
+            await send_str
+            r = await ws._response._session.get(
+                'https://api.coin.z.com/public/v1/status', auth=_Auth
+            )
+            data = await r.json()
+            before = datetime.datetime.fromisoformat(data['responsetime'][:-1])
+            while True:
+                await asyncio.sleep(1.0)
+                r = await ws._response._session.get(
+                    'https://api.coin.z.com/public/v1/status', auth=_Auth
+                )
+                data = await r.json()
+                after = datetime.datetime.fromisoformat(data['responsetime'][:-1])
+                delta = after - before
+                if delta.total_seconds() >= 1.0:
+                    break
+
+
+class RequestLimitHosts:
+    items = {
+        'api.coin.z.com': RequestLimit.gmocoin,
+    }
